mupdf_location = ''
pymupdf_version = '1.27.2.3'
pymupdf_version_tuple = (1, 27, 2, 3)
pymupdf_git_sha = ''
pymupdf_git_diff = ''
pymupdf_git_branch = ''
swig_version = '4.4.1'
swig_version_tuple = (4, 4, 1)
fake_no_gil = False
